/*
 * Programmieren fuer Physiker, SS 2023, Blatt 09, Aufgabe 23
 *
 * Klasse fuer rationale Zahlen, Rechenoperationen als Methoden
 */

#include <iostream>
#include <cstdlib>
using namespace std ;

// ------------------ GGT

long ggT (long a, long b) 
{
    a=abs(a) ;
    b=abs(b) ;
    while (b!=0)       // Euklidischer Algorithmus:
    {
	long h = a%b ; // h = Rest der Division von a%b
	a = b ;        // -> 0<=h<b, ggT(h,b) = ggT(a,b)
	b = h ;        // nimm als a nun b, nimm als b nun h.
    }
    return a ;
}

// ------------------ Class RATIO

class Ratio {
private:
    long zaehler, nenner ; // Zaehler und Nenner
    void kuerze();         // nur innerhalb Methoden aufrufbar
public:
    Ratio (long a, long b) ;  // Konstruktor
    Ratio operator+ (Ratio op2) ;
    Ratio operator- (Ratio op2) ;
    Ratio operator* (Ratio op2) ;    
    Ratio operator/ (Ratio op2) ;
    Ratio operator- () ;
    bool operator== (Ratio op2) ; // Zusatz: Gleichheit von Bruechen
    double toDouble () ;
    operator double() ;
    void print() ;
    // "<<" befreundet: Zugriff auf private Komponenten
    friend ostream & operator<< (ostream &, Ratio);  
} ;


Ratio::Ratio (long a=0, long b=1) // beachte die default Werte!
// es werden so auf einmal drei versch. Konstruktorvarianten definiert
{
    zaehler = a ;
    nenner = b ;
    if (0==nenner) cout << "Ooops. Illegal value assigned to a Ratio\n" ;
    kuerze() ;  // <- Mitgliedsfunktion wirkt auf aktuelles Objekt!
                // Aufruf innerhalb Mitgliedsfunktion ohne "." moeglich
}

void Ratio::kuerze()
{
    if (nenner<0) // Vorzeichen soll immer im Zaehler sein!
    { 
	nenner *= -1 ; 
	zaehler *= -1 ;
    } ;
    long g=ggT(zaehler,nenner) ;
    // cout << "Kuerze: zaehler, nenner, ggT: " 
    //      << zaehler << " " << nenner << " " << g << endl ;
    zaehler = zaehler / g ; 
    nenner = nenner / g ;
}

Ratio Ratio::operator+ (Ratio op2) 
{
    Ratio ergebn ;
    ergebn.zaehler = zaehler*op2.nenner + nenner*op2.zaehler ;
    ergebn.nenner = nenner*op2.nenner ;
    ergebn.kuerze() ;
    return ergebn ;
}

Ratio Ratio::operator- ()
{
    return Ratio(-zaehler, nenner) ; // Kurzvariante
}


Ratio Ratio::operator- (Ratio op2) 
{
    // Binaeres "-" hier als Beispiel aus unaerem "-"
    // und "+" zusammengebaut:
    Ratio ergebn = *this + (-op2) ; // "*this" ist aktuelles Objekt! 
    return ergebn ;
}
 
Ratio Ratio::operator* (Ratio op2) 
{
    Ratio ergebn ;
    ergebn.zaehler = zaehler*op2.zaehler ;
    ergebn.nenner = nenner*op2.nenner ;
    ergebn.kuerze() ;
    return ergebn ;
}   

Ratio Ratio::operator/ (Ratio op2) 
{
    // Hier mal eine kurze Variante:
    return Ratio (zaehler*op2.nenner, nenner*op2.zaehler) ;
}   

double Ratio::toDouble ()
{
    return double(zaehler)/double(nenner) ;
}

Ratio::operator double()
{
    return double(zaehler)/double(nenner);
}

void Ratio::print() 
{
    cout << "Der Bruch: " << zaehler << "/" << nenner << endl ;
}

ostream & operator<< (ostream & str, Ratio r)
{
    str << r.zaehler << "/" << r.nenner ;
    return str ;
}

bool Ratio::operator== (Ratio op2)
{
    if (zaehler != op2.zaehler) return false ;
    if (nenner != op2.nenner) return false ;
    return true ;
    // return (zaehler==op2.zaehler and nenner==op2.nenner) // <- Kurz!
}

// ------------------ MAIN program

int main()
{
    Ratio a1(2,15) ;
    Ratio a2(7,5) ;
    Ratio a3(3,7) ;
    Ratio a4(1,2) ;
    Ratio a5(1,3) ;
    Ratio ergebnis = a1/a2 + a3*(-a4+a5) ;
    cout << "Resultat: " ;
    ergebnis.print() ;
    cout << "Ergebnis als Bruch: " << ergebnis << endl ;
    cout << "Ergebnis als double: " << ergebnis.toDouble() << endl ;
    // cout << "Ergebnis als double: " << ergebnis.operator double() << endl ;
    // cout << "Ergebnis als double: " << (double)ergebnis << endl ;
    cout << "Ergebnis als double: " << double(ergebnis) << endl ;

    if (ergebnis==Ratio(2,84)) cout << "... Werte sind gleich!" << endl ; 

    // Kurzversion:
    cout << Ratio(2,15)/Ratio(7,5) + Ratio(3,7)*(-Ratio(1,2) + Ratio(1,3)) 
	 << endl ;
}
